// RSSInfo.cpp: implementation of the CRSSInfo class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RSSInfo.h"

unsigned __stdcall UpdateRSS(void* lpParameter);

#define LM_SETLABELTEXT 9600

typedef struct _SETLABELTEXTINFO
{
	string sName, sValue;
} SETLABELTEXTINFO, *LPSETLABELTEXTINFO;
BOOL CALLBACK LabelSearchProc(HWND hwnd, LPARAM lParam);

extern UINT_PTR LastEventIDValue;
extern HWND messageHandler;

////////////////////////////////////////////////////////////////////////////
//
// Construction
//
CRSSInfo::CRSSInfo(string sName, HINSTANCE hInstance)
{
	m_UpdateThread = 0;
	m_sName = sName;
}

////////////////////////////////////////////////////////////////////////////
//
// Destruction
//
CRSSInfo::~CRSSInfo()
{
	if (m_bActive)
	{
		KillTimer(messageHandler, m_uTimerID);
	}
	while (m_UpdateThread) Sleep(100);
}

////////////////////////////////////////////////////////////////////////////
//
// Update RSS variables
//
void CRSSInfo::Update()
{
	// if we are not on updating
	if (m_UpdateThread == 0)
	{
		_beginthreadex(NULL, 0, UpdateRSS, this, 0, &m_UpdateThread);
	}
}

////////////////////////////////////////////////////////////////////////////
//
// Read settings for the specified RSS feed
//
bool CRSSInfo::ReadSettings()
{
	char buf[MAX_LINE_LENGTH];
	m_Count				= GetRCInt(NAMED_OPTION(m_sName, "ItemCount"), 10);
	m_UpdateInterval	= GetRCInt(NAMED_OPTION(m_sName, "UpdateInterval"), 600);
	GetRCLine(NAMED_OPTION(m_sName, "URL"), buf, MAX_LINE_LENGTH, "http://www.shellfront.org/shellfront.xml");
	int len = strlen(buf);
	if (len > 0 && buf[len-1] == '\"')
		buf[len-1] = 0;
	m_sURL = (buf[0] == '\"') ? buf + 1 : buf;
	GetRCLine(NAMED_OPTION(m_sName, "InfoString"), buf, MAX_LINE_LENGTH, "^@Title^\n(^@Link^)\n^@Description^");
	m_sInfoString = buf;
	GetRCLine(NAMED_OPTION(m_sName, "ItemString"), buf, MAX_LINE_LENGTH, "^@Title^ (^@Link^)");
	m_sItemString = buf;
	GetRCLine(NAMED_OPTION(m_sName, "ItemSeparator"), buf, MAX_LINE_LENGTH, "\n");
	m_sItemSeparator = buf;
	GetRCLine(NAMED_OPTION(m_sName, "OnUpdated"), buf, MAX_LINE_LENGTH, "!none");
	m_sOnUpdated = buf;
	GetRCLine(NAMED_OPTION(m_sName, "OnFailure"), buf, MAX_LINE_LENGTH, "!none");
	m_sOnUpdateFailure = buf;
	GetRCLine(NAMED_OPTION(m_sName, "OnNewsUpdated"), buf, MAX_LINE_LENGTH, "!none");
	m_sOnNewsUpdated = buf;
	GetRCLine(NAMED_OPTION(m_sName, "InfoLabel"), buf, MAX_LINE_LENGTH, "");
	m_sTitleLabel = buf;
	GetRCLine(NAMED_OPTION(m_sName, "ItemLabel"), buf, MAX_LINE_LENGTH, "");
	m_sItemsLabel = buf;
	GetRCLine(NAMED_OPTION(m_sName, "LocalFile"), buf, MAX_LINE_LENGTH, "");
	if (buf[0] && buf[strlen(buf)-1] == '\"') buf[strlen(buf)-1] = 0;
	m_sLocalFile = (buf[0] == '\"') ?  buf+1 : buf;
	m_bActive = GetRCBool(NAMED_OPTION(m_sName, "Enabled"), TRUE) != FALSE;
	return true;
}

////////////////////////////////////////////////////////////////////////////
//
// Enable autoupdate
//
void CRSSInfo::Enable()
{
	if (m_bActive) return;
	m_bActive = true;
	LastEventIDValue++;
	m_nIDEvent = LastEventIDValue;
	m_uTimerID = SetTimer(messageHandler, m_nIDEvent, 1000*m_UpdateInterval, NULL);
}

////////////////////////////////////////////////////////////////////////////
//
// Disable autoupdate
//
void CRSSInfo::Disable()
{
	if (!m_bActive) return;
	m_bActive = false;
	KillTimer(messageHandler, m_uTimerID);
}

////////////////////////////////////////////////////////////////////////////
//
// Toggle autoupdate
//
void CRSSInfo::Toggle()
{
	m_bActive ? Disable() : Enable();
}

////////////////////////////////////////////////////////////////////////////
//
// Check pointer *pp for specified tag. If it is here, get a string marked
// by this tag, write it to the buffer and shift *pp
//
bool CheckFor(const char *tag, char **pp, char *buffer)
{
	if (strnicmp(*pp, tag, strlen(tag)) == 0)
	{
		*pp = strchr(*pp, '>');
		if (*pp == NULL)
			throw "Corrupt file";
		else
			(*pp)++;
		char *token = *pp;
		if (strnicmp(token, "<![cdata[", 9) == 0)
		{
			token += 9;
			*pp = strstr(token, "]]>");
			if (*pp != NULL)
			{
				**pp = 0;
				*pp += 3;
			}
		}
		if (*pp != NULL) *pp = strchr(*pp, '<');
		if (*pp == NULL)
			throw "Corrupt file";
		else
			**pp = 0;
		*pp = strchr((*pp)+1, '>');
		if (*pp == NULL)
			throw "Corrupt file";
		StringCchCopy(buffer, MAX_LINE_LENGTH, token);
		return true;
	} else return false;
}

////////////////////////////////////////////////////////////////////////////
//
// Substitute tag with szValue if possible
//
bool Substitute(char *szInfo, const char *szInfoString, const char *tag, const char *szValue, int &i)
{
	if (strnicmp(szInfoString + i, tag, strlen(tag)) == 0)
	{
		StringCchCat(szInfo, MAX_LINE_LENGTH, szValue);
		i += strlen(tag);
		return true;
	} else return false;
}

////////////////////////////////////////////////////////////////////////////
//
// Remove some HTML tags from the string
//
void DeHTML(char *tmp)
{
	char *str = new char[strlen(tmp)+1];
	int i=0, j=0;
	while (tmp[i])
	{
		if (strnicmp(tmp+i, "&lt;", 4) == 0)
		{
			str[j] = '<';
			j++;
			i+=4;
		} else if (strnicmp(tmp+i, "&gt;", 4) == 0)
		{
			str[j] = '>';
			j++;
			i+=4;
		} else if (strnicmp(tmp+i, "&raquo;", 7) == 0)
		{
			str[j] = '>'; str[j+1] = '>';
			j+=2;
			i+=7;
		}
		else if (strnicmp(tmp+i, "&amp;", 5) == 0)
		{
			str[j] = '&';
			j++;
			i+=5;
		} else if (strnicmp(tmp+i, "&quot;", 6) == 0)
		{
			str[j] = '\"';
			j++;
			i+=6;
		} else if (strnicmp(tmp+i, "<br>", 4) == 0)
		{
			str[j] = '\\'; str[j+1] = 'n';
			j+=2;
			i+=4;
		} else if (strnicmp(tmp+i, "<br/>", 5) == 0)
		{
			str[j] = '\\'; str[j+1] = 'n';
			j+=2;
			i+=5;
		} else if (strnicmp(tmp+i, "<p>", 3) == 0)
		{
			str[j] = '\\'; str[j+1] = 'n';
			j+=2;
			i+=3;
		} else if (tmp[i] == '<')
		{
			// skip all unknown tags
			do i++; while (tmp[i]!='>');
			i++;
		} else
		{
			str[j] = tmp[i];
			j++; i++;
		}
	}
	str[j] = 0;
	StringCchCopy(tmp, strlen(tmp)+1, str);
	delete str;
}

////////////////////////////////////////////////////////////////////////////
//
// Fill buffer using pattern and export RSS item data as LiteStep variable
//
void CRSSInfo::ExportRSSItemData(RSSItem *item,
					  const string *pattern,
					  char *buffer,
					  const char *szVarName)
{
	char szInfoString[MAX_LINE_LENGTH], _szInfoString[MAX_LINE_LENGTH];

	buffer[0] = 0;

	// trim doublequotes
	StringCchCopy(_szInfoString, MAX_LINE_LENGTH, pattern->c_str());
	int len = strlen(_szInfoString);
	if (_szInfoString[len-1] == '\"') _szInfoString[len-1] = 0;
	StringCchCopy(szInfoString, MAX_LINE_LENGTH, (_szInfoString[0] == '\"') ? _szInfoString+1 : _szInfoString);

	int i = 0, j = 0;
	while (szInfoString[i])
	{
		if (!(
			Substitute(buffer, szInfoString, "^@Title^",		item->title, i) ||
			Substitute(buffer, szInfoString, "^@Link^",			item->link, i) ||
			Substitute(buffer, szInfoString, "^@Description^",	item->description, i)
			))
		{
			buffer[j] = szInfoString[i];
			i++;
			j++;
			buffer[j] = 0;
		}
		else
		{
			j = strlen(buffer);
		}
	}
	SetVariable(szVarName, buffer, false);
}

////////////////////////////////////////////////////////////////////////////
//
// Read downloaded file with RSS news, parse it and fill variables
//
bool CRSSInfo::ReadFromFile(const char *szFile, bool bFailIfNotExist)
{
	bool bFailed = false;
	m_bNewsUpdated = false;

	HANDLE f = CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (f == INVALID_HANDLE_VALUE)
	{
		if (bFailIfNotExist)
			ErrorBox("Internal error: File successfully downloaded but cannot be open");
		return false;
	}
	DWORD rss_length = GetFileSize(f, NULL);
	char *buf = new char[rss_length];

	// Channel title, link and description

	RSSItem channel;
	ZeroMemory(&channel, sizeof(RSSItem));

	// let's start
	try
	{
		DWORD rd;
		if (!ReadFile(f, buf, rss_length, &rd, NULL))
			throw "Internal error: File successfully opened by cannot be read";

		// <?xml version="1.0" encoding="windows-1251" ?>
		char *p = strchr(buf, '<');
		if (p == NULL)
			throw "Recevied file is empty!";
		else
			p++;
		if (strnicmp(p, "?xml", 4) != 0)
		{
			throw "Received file's format differs from XML";
		}

		// "<rss version="0.91">
		p = strchr(p, '<');
		if (p == NULL)
			throw "This is not RSS file";
		else
			p++;
		
		// <channel>
		p = strchr(p, '<');
		if (p == NULL)
			throw "No channels found";
		else
			p++;

		// get header strings until <item> would be found
		for(;;)
		{
			p = strchr(p, '<');
			if (p == NULL)
				throw "No news items found";
			else
				p++;
			if (strnicmp(p, "item", 4) == 0) break;
			CheckFor("title", &p, channel.title) ||
			CheckFor("link",  &p, channel.link ) ||
			CheckFor("description", &p, channel.description);
		}
		DeHTML(channel.title);
		DeHTML(channel.link);
		DeHTML(channel.description);

		// now we shall process items
		vector<RSSItem*> items;
		RSSItem *item;
		for(;;)
		{
			if (items.size() == m_Count) break;
			if (strnicmp(p, "item", 4) != 0) break;
			p = strchr(p, '>');
			if (p == NULL)
				throw "<item> tag not finished. File is corrupt";
			else
				p++;
			item = new RSSItem;
			item->title[0] = 0;
			item->link[0] = 0;
			item->description[0] = 0;
			// fill item fields
			for(;;)
			{
				char *p1 = p;
				p = strchr(p, '<');
				if (p == NULL)
					throw "<item> group started but not finished. File is corrupt";
				else
					p++;
				if (strnicmp(p, "/item", 5) == 0) break;
				CheckFor("title", &p, item->title) ||
				CheckFor("link", &p, item->link) ||
				CheckFor("description", &p, item->description);
			}
			// skip </item>
			p = strchr(p, '>');
			if (p == NULL)
				throw "No </item> tag found";
			else
				p++;
			// find next <item> tag
			do
			{
				p = strchr(p, '<');
				if (p == NULL)
					throw "No </channel> tag found.";
				else
					p++;
			} while (strnicmp(p, "item", 4) != 0 && strnicmp(p, "/channel", 8) != 0);
			DeHTML(item->title);
			DeHTML(item->link);
			DeHTML(item->description);
			items.push_back(item);
		}

		// we finished!

		// fill channel title

		SetVariable(NAMED_OPTION(m_sName, "_Title"),  channel.title, false);
		SetVariable(NAMED_OPTION(m_sName, "_Link"),  channel.link, false);
		SetVariable(NAMED_OPTION(m_sName, "_Description"), channel.description, false);

		char szInfo[MAX_LINE_LENGTH];

		ExportRSSItemData(&channel, &m_sInfoString, szInfo, NAMED_OPTION(m_sName, "_Info"));

		m_sTitle = szInfo;

		// now let's fill items info

		char szItems[MAX_LINE_LENGTH];
		szItems[0] = 0;

		for (int i = 0; i<items.size(); i++)
		{
			char szItem[MAX_LINE_LENGTH];
			char szVarName[MAX_LINE_LENGTH];

			StringCchPrintf(szVarName, MAX_LINE_LENGTH, "%s_Item%d", m_sName.c_str(), i+1);
			ExportRSSItemData(items[i], &m_sItemString, szItem, szVarName);
			
			StringCchPrintf(szVarName, MAX_LINE_LENGTH, "%s_Item%d_Title", m_sName.c_str(), i+1);
			SetVariable(szVarName,  items[i]->title, true);
			StringCchPrintf(szVarName, MAX_LINE_LENGTH, "%s_Item%d_Link", m_sName.c_str(), i+1);
			SetVariable(szVarName,  items[i]->link, true);
			StringCchPrintf(szVarName, MAX_LINE_LENGTH, "%s_Item%d_Description", m_sName.c_str(), i+1);
			SetVariable(szVarName,  items[i]->description, false);

			if (i > 0)
				StringCchCat(szItems, MAX_LINE_LENGTH, m_sItemSeparator.c_str());
			StringCchCat(szItems, MAX_LINE_LENGTH, szItem);
		}
		m_sItems = szItems;
		SetVariable(NAMED_OPTION(m_sName, "_Items"), szItems, false);

		vector<RSSItem*>::iterator ri = items.begin(), rend = items.end();
		for (; ri != rend; ri++) delete (*ri);
		items.clear();

	} catch (const char *msg)
	{
		// one of our error messages
		char errmsg[MAX_LINE_LENGTH];
		StringCchPrintf(errmsg, MAX_LINE_LENGTH, "The following error occured during RSS file parsing:\n%s", msg);
		ErrorBox(errmsg);
		bFailed = true;
	}
	catch (...)
	{
		// unspecified exception
		ErrorBox("Unknown error occured during RSS file parsing.");
		bFailed = true;
	}
	// let's stop :)

	delete buf;

	// just close file handle, caller would erase file by itself
	CloseHandle(f);
	return !bFailed;
}


////////////////////////////////////////////////////////////////////////////
//
// Titles, description, etc. can contain spaces...
// Also we shall monitor changes here
//
void CRSSInfo::SetVariable(const char *szEvar, const char *szValue, bool bTrackChanges)
{
	char nv[MAX_LINE_LENGTH] = "\"";
	char szOldValue[MAX_LINE_LENGTH];

	StringCchCat(nv, MAX_LINE_LENGTH, szValue);

	// all double quotes would be changed to single quotes...
	// sorry, but LS is a real hell...
	for(int i=1; nv[i]; i++)
	{
		if (nv[i] == '\"') nv[i] = '\'';
	}
	StringCchCat(nv, MAX_LINE_LENGTH, "\"");

	if (bTrackChanges)
	{
		if (!LSGetVariableEx(szEvar, szOldValue, MAX_LINE_LENGTH))
			m_bNewsUpdated = true;
		else
		{
			if (strncmp(szOldValue, nv+1, strlen(nv)-2) != 0)
			{
				m_bNewsUpdated = true;
			}
		}
	}
	LSSetVariable(szEvar, nv);
}



////////////////////////////////////////////////////////////////////////////
//
// Set text for the Label, xLabel or xLabelLight window.
// 
void SetLabelText(string sLabelName, string sValue)
{
	static HWND hwndLitestep = GetLitestepWnd();

	SETLABELTEXTINFO slti;
	slti.sName = sLabelName;
	slti.sValue = sValue;
	EnumWindows(LabelSearchProc, (LPARAM)&slti);
}

////////////////////////////////////////////////////////////////////////////
//
// Enumeration
// 
BOOL CALLBACK LabelSearchProc(HWND hwnd, LPARAM lParam)
{
	char buf[MAX_LINE_LENGTH], buf2[MAX_LINE_LENGTH];
	if (GetWindowText(hwnd, buf, MAX_LINE_LENGTH))
	{
		if (buf == ((LPSETLABELTEXTINFO)lParam)->sName )
		{
			if (GetClassName(hwnd, buf2, MAX_LINE_LENGTH))
			{
				if (stricmp(buf2,"xLabel") == 0 || stricmp(buf2, "LabelLS") == 0)
				{
					SendMessage(hwnd, LM_SETLABELTEXT, 0, (LPARAM)((LPSETLABELTEXTINFO)lParam)->sValue.c_str());
					return FALSE;
				}
			}
		}
	}
	EnumChildWindows(hwnd, LabelSearchProc, lParam);
	return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//
// This thread checks RSS news.
// We cannot download file from the main thread cause it can freeze LS
//
//DWORD WINAPI UpdateRSS(LPVOID lpParameter)
unsigned __stdcall UpdateRSS(void* lpParameter)
{
	CRSSInfo *rss = (CRSSInfo*)lpParameter;
	char tempfilename[MAX_LINE_LENGTH] = "";
	try
	{
		if (rss->m_sLocalFile != "")
			StringCchCopy(tempfilename, MAX_LINE_LENGTH, rss->m_sLocalFile.c_str());
		else
		{
			GetTempPath(MAX_LINE_LENGTH, tempfilename);
			StringCchCat(tempfilename, MAX_LINE_LENGTH, "\\lsrss.xml");
		}
		if (URLDownloadToFile(NULL, rss->m_sURL.c_str(), tempfilename, 0, NULL) != S_OK)
			throw "There is no file with specified URL";
		if (!rss->ReadFromFile(tempfilename, true)) throw "RSS file parsing failed";
		if (stricmp(rss->m_sOnUpdated.c_str(), "!none") != 0)
		{
			LSExecute(NULL, rss->m_sOnUpdated.c_str(), SW_SHOWNORMAL);
			if (rss->m_bNewsUpdated &&
				stricmp(rss->m_sOnNewsUpdated.c_str(), "!none") != 0 &&
				rss->m_sLocalFile != "")
			{
				rss->m_bNewsUpdated = false;
				LSExecute(NULL, rss->m_sOnNewsUpdated.c_str(), SW_SHOWNORMAL);
			}
		}
		if (rss->m_sTitleLabel != "")
			SetLabelText(rss->m_sTitleLabel, rss->m_sTitle);
		if (rss->m_sItemsLabel != "")
			SetLabelText(rss->m_sItemsLabel, rss->m_sItems);
	}
	catch(const char *)
	{
		if (stricmp(rss->m_sOnUpdateFailure.c_str(), "!none") != 0)
		{
			LSExecute(NULL, rss->m_sOnUpdateFailure.c_str(), SW_SHOWNORMAL);
		}
	}
	catch(...)
	{
		ErrorBox("Internal error: Unknown exception raised");
	}
	rss->m_UpdateThread = 0;
	return 1;
}
